// SCR v.2.0 от 18.04.2019

#include <LCD_1602_RUS.h>                                               // Библиотека для работы с LCD 1602 с русскими символами
#include <EEPROM.h>                                                     // Библиотека для работы с EEPROM
#include <Wire.h>                                                       // Библиотека для работы с шиной I2C

//------------------------------------ Настройка подключения пинов к ардуино

#define detect      3                                                   // Детектор нуля
#define triac       5                                                   // Управление симистором
#define SW          7                                                   // Энкодер SW кнопка
#define CLK         8                                                   // Энкодер CLK
#define DT          9                                                   // Энкодер DT
#define ACCELERATOR 12                                                  // Вывод на максимальную, разгонную мощность (подтянут вверх, активен когда замкнут на GND, при отпускании вызывает мощность из памяти)
#define STOP        13                                                  // Вывод сброс мощности в нуль (подтянут вверх, активен когда замкнут на GND, имеет приоритет над разгоном)
#define button1     10                                                  // Кнопка 1 "Режим настройки"
#define button2     11                                                  // Кнопка 2 "ВКЛ/ВЫКЛ"

LCD_1602_RUS lcd(0x3F, 16, 2);                                          // Адрес для подключения LCD экрана через I2C lcd(0x27, 16, 2);

//------------------------------------ Константы

double resist_ten = 18.30;                                              // Cопротивление ТЭНа (3 кВт), Ом (ввести измеренное значение)
double TT_coeff = 0.0604788;                                            // Коэффициент ТТ |20 А (ввести откорректированное значение)
int max_W = 3000;                                                       // Максимальная мощность, Вт
int step_W = 1;                                                         // Шаг установки мощности, Вт

//------------------------------------ Переменные настроек

int preset_max = 4;                                                     // Число предустановок + 1 (0) = 5
char* name_preset[] = {"РКТ ", "ДСТ ", "НБК ", "Св1 ", "Св2 "};
String Var;
String T1, T2;
int ust_W = 0;                                                          // Расчетная мощность, Вт
double ust_I = 0;                                                       // Установленный ток, А
int buffer_ust_W[10];                                                   // Сохраненная в EEPROM установленная мощность
int angle_I = 2250;                                                     // Угол открытия симистора
int angle_I_ocr = 2250;
int preset = 0;                                                         // Флаг меню предустановок

//------------------------------------ Переменные для расчета RMS тока и мощности

volatile unsigned long Iism = 0;                                        // Мгновенные значения тока
volatile int cntr = 0;                                                  // Счетчик в обработчике прерывания
volatile unsigned long Isumm = 0;                                       // Сумма квадратов тока
double real_I = 0;                                                      // Расчетный RMS ток, А
int real_W = 0;                                                         // Расчетная мощность, Вт
double sqrt_I_sum = 0;                                                  // Переменная для расчета квадратного корня
int zero = 0;                                                           // Детект нуля

//------------------------------------ Переменные для таймингов

unsigned long previousMillis_info = 0;                                  // Момент времени для передачи информации в последовательный порт 9600
unsigned long displayMillis = 0;                                        // Момент времени для обновления LCD
unsigned long countMillis = 0;                                          // Момент времени для регулирования тока(мощности)
unsigned long encoderMillis = 0;                                        // Момент времени для энкодера

//------------------------------------ Переменные для энкодера

int encoder_CLK = 0;                                                    // Обработка вращения
int encoder_DT = 0;                                                     // Обработка вращения
int encoder_CLK_prev = 0;                                               // Обработка вращения
byte set_menu_W = false;                                                // Обработка нажатия
unsigned long SW_time_press = 0;                                        // Момент времени нажатия
unsigned long SW_press_time = 0;                                        // Счетчик времени нажатия

//------------------------------------ Переменная разгонной мощности

byte set = 0;                                                           // Переменная включения разгонной, максимальной мощности
byte ACCELERATOR_press = 0;
byte STOP_press = 0;
unsigned long ACCELERATOR_time_press = 0;                               // Момент времени нажатия
unsigned long ACCELERATOR_press_time = 0;                               // Счетчик времени нажатия
unsigned long STOP_time_press = 0;                                      // Момент времени нажатия
unsigned long STOP_press_time = 0;                                      // Счетчик времени нажатия

//------------------------------------ Переменные для Кнопок 1 и 2

byte set_I = 0;                                                         // Переменная Кнопки 2 "Вкл/Выкл"
byte button1_press = 0;
byte button2_press = 0;
unsigned long button1_time_press = 0;                                   // Момент времени нажатия
unsigned long button1_press_time = 0;                                   // Счетчик времени нажатия
unsigned long button2_time_press = 0;                                   // Момент времени нажатия
unsigned long button2_press_time = 0;                                   // Счетчик времени нажатия

//------------------------------------ Процедура инициализации

void setup(void) {

  //------------------------------------ Инициализация Serial port 9600  

  Serial.begin(9600);                                                   // Инициализация приема сообщения в Serial port 9600

  //------------------------------------ Инициализация вводов-выводов

  pinMode(detect, INPUT);                                               // Инициализация входа детекта нуля
  pinMode(triac, OUTPUT);                                               // Инициализация выхода управления симистором
  pinMode(CLK, INPUT);                                                  // Подтягивание резистора входа энкодера CLK
  pinMode(DT, INPUT);                                                   // Подтягивание резистора входа энкодера DT
  pinMode(SW, INPUT);                                                   // Подтягивание резистора входа энкодера SW
  pinMode(button1, INPUT);                                              // Подтягивание резистора входа Кнопка 1
  pinMode(button2, INPUT);                                              // Подтягивание резистора входа Кнопка 2
  pinMode(ACCELERATOR, INPUT);                                          // Подтягивание резистора входа Разгонная, максимальная мощность
  pinMode(STOP, INPUT);                                                 // Подтягивание резистора входа Сброс мощности в нуль

  digitalWrite(CLK, HIGH);                                              // Включаем подтягивающий резистор
  digitalWrite(DT, HIGH);                                               // Включаем подтягивающий резистор
  digitalWrite(SW, HIGH);                                               // Включаем подтягивающий резистор  
  digitalWrite(button1, HIGH);                                          // Включаем подтягивающий резистор
  digitalWrite(button2, HIGH);                                          // Включаем подтягивающий резистор
  digitalWrite(ACCELERATOR, HIGH);                                      // Включаем подтягивающий резистор
  digitalWrite(STOP, HIGH);                                             // Включаем подтягивающий резистор

  //------------------------------------ Инициализация LCD

  lcd.init();
  lcd.backlight();
  lcd.setCursor(5, 0);
  lcd.print("СРМ-ТТ");
  lcd.setCursor(3, 1);
  lcd.print("Версия 2.0");                                              // Изменения от 18.04.19
  delay(1000);
  lcd.clear();

  //------------------------------------ Считывание настроек из EEPROM

  for (preset = preset_max ; preset >= 0 ; preset--) {
    buffer_ust_W[preset] = EEPROM_read(preset * 2);
  }
  preset = 0;

  //------------------------------------ Настройка АЦП  

  ADMUX = (0 << REFS1) | (1 << REFS0) | (0 << MUX2) | (0 << MUX1) | (1 << MUX0);
  ADCSRA = B11101111;
  ACSR = (1 << ACD);
  TCCR1A = 0x00;
  TCCR1B = 0x00;
  TCCR1B = (0 << CS12) | (1 << CS11) | (1 << CS10);
  OCR1A = 0;
  TIMSK1 = 0x00;

  //------------------------------------ Вызов прерывания при детектировании нуля

  attachInterrupt(1, zero_crosss_int, RISING);                          // вызов прерывания при детектировании нуля

}

//------------------------------------  Прерывание по нулю

void zero_crosss_int() {
  TCNT1 = 0;
  OCR1A = angle_I_ocr;
  zero++;
}

//------------------------------------ Обработка таймера по совпадению нуля

ISR (TIMER1_COMPA_vect) {
  PORTD |=  (1 << PORTD5);
  TCNT1 = 65535 - 3;
}

ISR (TIMER1_OVF_vect) {
  PORTD &=  ~(1 << PORTD5); 
  TCNT1 = OCR1A + 1;
}

//------------------------------------ Обработка прерывания АЦП для расчета среднеквадратичного тока

ISR(ADC_vect) {
  byte An_pin = ADCL;
  byte An = ADCH;
  Iism = ((An << 8) + An_pin) - 512;
  Iism *= Iism;                                                         // возводим значение в квадрат
  Isumm += Iism;                                                        // складываем квадраты измерений
  cntr++;
}

//------------------------------------ Считывание из EEPROM

int EEPROM_read(int addr) {
  byte raw[2];
  for (byte i = 0; i < 2; i++) raw[i] = EEPROM.read(addr + i);
  int &num = (int&)raw;
  return num;
}

//------------------------------------ Запись в EEPROM

void EEPROM_write(int addr, int num) {
  byte raw[2];
  (int&)raw = num;
  for (byte i = 0; i < 2; i++) EEPROM.update(addr + i, raw[i]);
}

void parseParams(String inputString) {                                  // Функция разбора принятой в Serial port строки
  String bb = inputString;
  if (bb.substring(0, 2) == "TW") {
    T1 = bb.substring(bb.indexOf("TW", 2) + 3);
    ust_I = T1.toFloat();                                               // Выставленный ток с Serial порта
    buffer_ust_W[preset] = ust_I;
  }
}

void loop(void) {

  //------------------------------------ Считывание с Serial port

  while (Serial.available()) {                                          // Serial port, пока не конец сообщения, читаем данные и формируем строку
    char ch = Serial.read();
    Var += ch;
    if (ch == '\n') {
      parseParams(Var);                                                 // вызываем функцию разбора принятой строки
      Var = "";
    }
  }

  //------------------------------------ Расчет RMS тока и мощности

  if (zero >= 6) {                                                      // zero - кколичество полупериодов для рассчета среднеквадратичного значения тока
    sqrt_I_sum = Isumm / cntr;
    real_I = sqrt(sqrt_I_sum) * TT_coeff;
    real_W = real_I * real_I * resist_ten;
    cntr = 0;
    Isumm = 0;
    zero = 0;    
  }

  //------------------------------------ Регулятор тока

  ust_I = sqrt(buffer_ust_W[preset] / resist_ten);
  
  if (millis() - countMillis >= 15) {                                   // Если регулировка долгая, то уменьшите значение, если видно частые изменения регулировки то увеличте значение
    countMillis = millis();
    if (real_I > ust_I + 0.05) {
      angle_I++;
    }
    if (real_I < ust_I - 0.05) {
      angle_I--;
    }
    if (real_I < ust_I + 0.05 && real_I > ust_I - 0.05) {
      angle_I = angle_I;
    }
    if (angle_I > 2250) {
      angle_I = 2250;
    }
    if (angle_I <= 1) {
      angle_I = 1;
    }
  }

  if (ust_W < 0 || buffer_ust_W[preset] < 0) {                          // Если установленное значение меньше 0, то 0
    ust_W = 0;
    buffer_ust_W[preset] = 0;
  }

  if (set_I) {
    ust_W = (buffer_ust_W[preset] > 0) ? buffer_ust_W[preset] : 0;
    TIMSK1 |= (1 << OCIE1A) | (1 << TOIE1);
  }
  else
  {
    ust_I = 0;
    TIMSK1 = 0x00;
    PORTD &=  ~(1 << PORTD5);
  }

  angle_I_ocr = angle_I;

  //------------------------------------ Обработка входа "Сброс мощности"

   if (digitalRead(STOP) == LOW ) {                                       // Счётчик времени активного состояния входа "Сброс мощности"
    if (millis() - STOP_time_press > 1) {
      STOP_press_time++;
      STOP_time_press = millis();
    }
  }
  
  if (STOP_press_time > 500) {                                          // Если вход "Сброс мощности" активен более ХХХ мс, то мощность устанавливается в ноль
    STOP_press_time = 0;
    buffer_ust_W[preset] = 0;
    set = 3;
    }
    
  if (digitalRead(STOP) == HIGH && set == 3) {                          // Сброс счетчика, если вход "Сброс мощности" не активен, выход на заданную мощность
    buffer_ust_W[preset] = EEPROM_read(preset * 2);
    set = 0;
    lcd.clear();
  }

  //------------------------------------ Обработка входа "Разгонная, максимальная мощность"

  if (digitalRead(ACCELERATOR) == LOW ) {                               // Счётчик времени активного состояния входа "Разгонная, максимальная мощность"
    if (millis() - ACCELERATOR_time_press > 1) {
      ACCELERATOR_press_time++;
      ACCELERATOR_time_press = millis();
    }
  }
  
  if (digitalRead(STOP) == HIGH && ACCELERATOR_press_time > 500) {      // Если не активен вход "Сброс мощности" и вход "Разгонная, максимальная мощность" активен более ХХХ мс, то мощность устанавливается максимальной
    ACCELERATOR_press_time = 0;
    buffer_ust_W[preset] = max_W;
    set = 1;
  }

  if (digitalRead(ACCELERATOR) == HIGH && set == 1) {                   // Сброс счетчика, если вход "Разгонная, максимальная мощность" не активен и перед этим был включен, выход на заданную мощность
    buffer_ust_W[preset] = EEPROM_read(preset * 2);
    set = 0;
    lcd.clear();
  }

  //------------------------------------ Обработка Кнопки 1 (Режим настройки)

  if (digitalRead(button1) == LOW) {                                    // Счётчик времени активного состояния Кнопки 1
    if (millis() - button1_time_press > 1) {
      button1_press_time++;
      button1_time_press = millis();
    }
  }

  if (digitalRead(button1) == HIGH) {                                   // Сброс счетчика, если Кнопка 1 не активна
    button1_press_time = 0;
  }

  if ( button1_press_time >= 200) {                                     // Если кнопка активна более ХХХ мс, то - режим установки
    if (!set_menu_W) {//
      set_menu_W = true;
    }
    else {
      set_menu_W = false;
    }
    button1_press_time = 0;
    lcd.clear();
  }

  //------------------------------------ Обработка Кнопки 2 (ВКЛ/ВЫКЛ регулятора)

  if (digitalRead(button2) == LOW) {                                    // Счётчик времени активного состояния Кнопки 2
    if (millis() - button2_time_press > 1) {
      button2_press_time++;
      button2_time_press = millis();
    }
  }

  if (digitalRead(button2) == HIGH) {                                   // Сброс счетчика, если Кнопка 2 не активна
    button2_press_time = 0;
  }

  if ( button2_press_time >= 200 && !set_menu_W) {                      // Если кнопка активна более ХХХ мс, то - режим включения/выключения
    if (!set_I) {//
      set_I = true;
    }
    else {
      set_I = false;
    }
    button2_press_time = 0;
    lcd.clear();
  }

  //------------------------------------ Обработка кнопки энкодера

  if (digitalRead(SW) == LOW) {                                         // Счётчик времени активного состояния SW
    if (millis() - SW_time_press > 1) {
      SW_press_time++;
      SW_time_press = millis();
    }
  }

  if (digitalRead(SW) == HIGH) {                                        // Сброс счетчика, если SW не активна
    SW_press_time = 0;
  }

  if ( SW_press_time >= 200 && set_menu_W) {                            // Если SW нажата более ХХХ мс, то запись в память
    SW_press_time = 0;
    EEPROM_write(preset * 2, buffer_ust_W[preset]);
    lcd.setCursor(10, 0);
    lcd.print("  СОХР");
    delay(1000);
    lcd.clear();
  }

  //------------------------------------ Меню установки мощности
  
  if (set_menu_W) {
    if (millis() >= (encoderMillis + 5)) {
      encoder_CLK = digitalRead(CLK);
      encoder_DT = digitalRead(DT);
      if ((!encoder_CLK) && (encoder_CLK_prev)) { 
        if (encoder_DT) {
          if (buffer_ust_W[preset] <= max_W - step_W) buffer_ust_W[preset] = buffer_ust_W[preset] + step_W;
        }
        else {
          if (buffer_ust_W[preset] >= 0 + step_W) buffer_ust_W[preset] = buffer_ust_W[preset] - step_W;
        }
      }
      encoder_CLK_prev = encoder_CLK;
      encoderMillis = millis();
    }
  }

  //------------------------------------ Прокрутка настроек  

  if (!set_menu_W) {
    if (millis() >= (encoderMillis + 5)) {
      encoder_CLK = digitalRead(CLK);
      encoder_DT = digitalRead(DT);
      if ((!encoder_CLK) && (encoder_CLK_prev)) {
        if (encoder_DT) {
          preset = preset + 1;
          if (preset > preset_max) preset = 0;
        }
        else {
          preset = preset - 1;
          if (preset < 0) preset = preset_max;
        }
      }
      encoder_CLK_prev = encoder_CLK;
      encoderMillis = millis();
    }
  }

  //------------------------------------ Вывод информации на LCD

  if (millis() - displayMillis >= 300) {
    displayMillis = millis();

  //--- Первая строка
  
  if (!set_menu_W) {
      lcd.setCursor(0, 0);
      lcd.print(name_preset[preset]);
             if (set==3) {
             lcd.setCursor(4, 0);
             lcd.print(">СТОП<");
             }
             else if (set==1) {
             lcd.setCursor(4, 0);
             lcd.print(">РЗГН<");             
             }
             else {
             lcd.print(buffer_ust_W[preset]);
             lcd.print("Вт ");
             }
  if (set_I) {
      lcd.setCursor(11, 0);
      lcd.print(">ВКЛ<");
      }
      else {
      lcd.setCursor(11, 0);
      lcd.print(">ВЫКЛ");
      }
  
  //--- Вторая строка

      lcd.setCursor(0, 1);
      lcd.print("Изм ");
      lcd.print(real_W);
      lcd.print("Вт ");
      lcd.setCursor(11, 1);
      lcd.print(real_I);
      lcd.setCursor(15, 1);
      lcd.print("А");
  }
 
  //--- Меню уставок
  
    if (set_menu_W) {
      lcd.setCursor(0, 0);
      lcd.print("УСТАНОВКА:");
      lcd.setCursor(0, 1);
      lcd.print(name_preset[preset]);
      lcd.print(">> ");
      lcd.print(buffer_ust_W[preset]);
      lcd.print("Вт    ");
    }
  }
  
  //------------------------------------ Вывод информации в порт 9600

  if (millis() - previousMillis_info >= 1000) {
    previousMillis_info = millis();

    Serial.print("realI ");                                             // Измеренный среднеквадратичный ток
    Serial.println(real_I);

    Serial.print("ustI ");                                              // Установленный выходной ток
    Serial.println(ust_I);

    Serial.print("angleI ");                                            // Угол установленный
    Serial.println(angle_I);

    Serial.println("-----------------");
  }
}
